///////////////////////////////////////////////////////////////////////////////
//
/// \file       alone_decoder.c
/// \brief      Decoder for LZMA_Alone files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "alone_common.h"
#include "lzma_encoder.h"


static lzma_ret
alone_encode(lzma_coder *coder,
		lzma_allocator *allocator lzma_attribute((unused)),
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size,
		lzma_action action)
{
	while (*out_pos < out_size)
	switch (coder->sequence) {
	case SEQ_PROPERTIES:
		if (lzma_lzma_encode_properties(
				&coder->options, out + *out_pos)) {
			coder->sequence = SEQ_UNINITIALIZED;
			return LZMA_HEADER_ERROR;
		}

		coder->sequence = SEQ_DICTIONARY_SIZE;
		++*out_pos;
		break;

	case SEQ_DICTIONARY_SIZE:
		out[*out_pos] = coder->options.dictionary_size
				>> (coder->pos * 8);

		if (++coder->pos == 4) {
			coder->pos = 0;
			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
		}

		++*out_pos;
		break;

	case SEQ_UNCOMPRESSED_SIZE:
		out[*out_pos] = coder->uncompressed_size
				>> (coder->pos * 8);

		if (++coder->pos == 8) {
			coder->pos = 0;
			coder->sequence = SEQ_CODER_INIT;
		}

		++*out_pos;
		break;

	case SEQ_CODER_INIT: {
		// Two is enough because there won't be implicit filters.
		lzma_filter_info filters[2] = {
			{
				.init = &lzma_lzma_encoder_init,
				.options = &coder->options,
				.uncompressed_size = coder->uncompressed_size,
			}, {
				.init = NULL,
			}
		};

		const lzma_ret ret = lzma_next_coder_init(&coder->next,
				allocator, filters);

		if (ret != LZMA_OK) {
			coder->sequence = SEQ_UNINITIALIZED;
			return ret;
		}

		coder->sequence = SEQ_CODE;
	}

	// Fall through

	case SEQ_CODE: {
		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		switch (ret) {
		case LZMA_OK:
			assert(*in_pos == in_size || *out_pos == out_size);
			return LZMA_OK;

		case LZMA_STREAM_END:
			coder->sequence = SEQ_END;
			return LZMA_STREAM_END;

		default:
			assert(ret == LZMA_DATA_ERROR);
			coder->sequence = SEQ_UNINITIALIZED;
			return ret;
		}

		break;
	}

	case SEQ_END:
		return LZMA_STREAM_END;

	case SEQ_UNINITIALIZED:
	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
alone_encoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_alone_encoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_options_alone *options)
{
	if (next->init != &lzma_alone_encoder_init) {
		// Get rid of the previous coder.
		lzma_next_coder_end(next, allocator);

		// Allocate data structure for LZMA_Alone coder.
		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		// Mark that we don't have any LZMA coder initialized yet.
		next->coder->next = LZMA_NEXT_CODER_INIT;
	}

	// Initialize the LZMA_Alone coder variables.
	next->coder->sequence = SEQ_PROPERTIES;
	next->coder->pos = 0;
	next->coder->options = options->lzma;
	next->coder->uncompressed_size = options->uncompressed_size;

	// Initialization successful, set the function pointers.
	next->init = &lzma_alone_encoder_init;
	next->code = &alone_encode;
	next->end = &alone_encoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_alone_encoder(lzma_stream *strm, const lzma_options_alone *options)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	ret = lzma_alone_encoder_init(
			&strm->internal->next, strm->allocator, options);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
