///////////////////////////////////////////////////////////////////////////////
//
/// \file       block_decoder.c
/// \brief      Decodes .lzma Blocks
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "block_decoder.h"
#include "raw_decoder.h"
#include "check.h"


struct lzma_coder_s {
	/// The filters in the chain; initialized with lzma_raw_decoder_init().
	lzma_next_coder next;

	enum {
		SEQ_HEADER,
		SEQ_CODER_INIT,
		SEQ_CODE,
		SEQ_CHECK,
	} sequence;

	/// Position in variable-length integers (and in some other places).
	size_t pos;

	/// Check of the uncompressed data
	lzma_check check;

	/// The real Compressed Size as calculated while decoding.
	uint64_t compressed_size;

	/// The real Uncompressed Size as calculated while decoding.
	uint64_t uncompressed_size;

	lzma_next_coder block_header_decoder;
	lzma_options_block options;
};


static lzma_ret
block_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size && (*in_pos < in_size
			|| coder->sequence == SEQ_CODE
			|| coder->sequence == SEQ_CODER_INIT))
	switch (coder->sequence) {
	case SEQ_HEADER: {
		const lzma_ret ret = coder->block_header_decoder.code(
				coder->block_header_decoder.coder,
				allocator, in, in_pos, in_size,
				NULL, NULL, 0, LZMA_RUN);
		if (ret != LZMA_STREAM_END)
			return ret;

		// If we knew Compressed or Uncompressed Size, verify that
		// they match the values possibly stored in the Block Header.
		if (coder->compressed_size != UINT64_MAX) {
			if (coder->options.compressed_size != UINT64_MAX
					&& coder->options.compressed_size
					!= coder->compressed_size)
				return LZMA_DATA_ERROR;

			// Since the Block decoder was told the Compressed
			// Size, store it to the options structure. This
			// way we can verify the Compressed Size given to
			// even if the Compressed Size field wasn't present
			// in the Block Header.
			coder->options.compressed_size
					= coder->compressed_size;
		}

		// Same things for Uncompressed Size.
		if (coder->uncompressed_size != UINT64_MAX) {
			if (coder->options.uncompressed_size != UINT64_MAX
					&& coder->options.uncompressed_size
					!= coder->uncompressed_size)
				return LZMA_DATA_ERROR;

			coder->options.uncompressed_size
					= coder->uncompressed_size;
		}

		// Reset the sizes. They will contain the real sizes,
		// which get calculated while decoding.
		coder->compressed_size = 0;
		coder->uncompressed_size = 0;

		coder->sequence = SEQ_CODER_INIT;
		break;
	}

	case SEQ_CODER_INIT: {
		const lzma_ret ret = lzma_raw_decoder_init(&coder->next,
				allocator, coder->options.filters,
				coder->options.has_eopm
					? UINT64_MAX
					: coder->options.uncompressed_size,
					true);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_CODE;
	}

	// Fall through

	case SEQ_CODE: {
		const size_t in_start = *in_pos;
		const size_t out_start = *out_pos;

		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		(void)lzma_check_update(&coder->check, coder->options.check,
				out + out_start, *out_pos - out_start);

		coder->compressed_size += *in_pos - in_start;
		coder->uncompressed_size += *out_pos - out_start;

		// Verify that the sizes stay in the valid range.
		if (coder->compressed_size > coder->options.compressed_size
				|| coder->uncompressed_size
					> coder->options.uncompressed_size)
			return LZMA_DATA_ERROR;

		if (ret != LZMA_STREAM_END)
			return ret;

		// If Compressed Size was known, it must match now.
		if (coder->options.compressed_size != UINT64_MAX
				&& coder->compressed_size
					!= coder->options.compressed_size)
			return LZMA_DATA_ERROR;

		// If Uncompressed Size was known, it must match now.
		if (coder->options.uncompressed_size != UINT64_MAX
				&& coder->uncompressed_size
					!= coder->options.uncompressed_size)
			return LZMA_DATA_ERROR;

		lzma_check_finish(&coder->check, coder->options.check);
		coder->sequence = SEQ_CHECK;
		break;
	}

	case SEQ_CHECK:
		switch (coder->options.check) {
		case LZMA_CHECK_NONE:
			return LZMA_STREAM_END;

		case LZMA_CHECK_CRC32:
			if (((coder->check.crc32 >> (coder->pos * 8))
					& 0xFF) != in[*in_pos])
				return LZMA_DATA_ERROR;
			break;

		case LZMA_CHECK_CRC64:
			if (((coder->check.crc64 >> (coder->pos * 8))
					& 0xFF) != in[*in_pos])
				return LZMA_DATA_ERROR;
			break;

		case LZMA_CHECK_SHA256:
			if (coder->check.sha256.buffer[coder->pos]
					!= in[*in_pos])
				return LZMA_DATA_ERROR;
			break;

		default:
			assert(coder->options.check <= 7);
			break;
		}

		++*in_pos;

		if (++coder->pos == lzma_check_sizes[coder->options.check])
			return LZMA_STREAM_END;

		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
block_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);

	// The eighth one is never used.
	for (size_t i = 0; i < 7; ++i)
		lzma_free(coder->options.filters[i].options, allocator);

	lzma_free(coder, allocator);

	return;
}


extern lzma_ret
lzma_block_decoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_options_block *options)
{
	if (next->init != &lzma_block_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		next->coder->next = LZMA_NEXT_CODER_INIT;
		next->coder->block_header_decoder = LZMA_NEXT_CODER_INIT;

		// The eighth one is never used for any data, but better
		// set it to NULL because it simplifies code a bit later.
		for (size_t i = 0; i < 8; ++i)
			next->coder->options.filters[i].options = NULL;
	}

	// Validate the given options.
	if (options->check > LZMA_CHECK_ID_MAX
			|| !lzma_vli_is_valid(options->uncompressed_size)
			|| !lzma_vli_is_valid(options->compressed_size)) {
		block_decoder_end(next->coder, allocator);
		return LZMA_STREAM_ERROR;
	}

	next->coder->sequence = SEQ_HEADER;
	next->coder->pos = 0;

	next->coder->uncompressed_size = options->uncompressed_size;
	next->coder->compressed_size = options->compressed_size;

	next->coder->options.check = options->check;
	next->coder->options.has_crc32 = options->has_crc32;

	for (size_t i = 0; i < 8; ++i) {
		lzma_free(next->coder->options.filters[i].options, allocator);
		next->coder->options.filters[i].options = NULL;
	}

	{
		const lzma_ret ret = lzma_block_header_decoder_init(
				&next->coder->block_header_decoder, allocator,
				&next->coder->options);
		if (ret != LZMA_OK) {
			block_decoder_end(next->coder, allocator);
			return ret;
		}
	}

	next->init = &lzma_block_decoder_init;
	next->code = &block_decode;
	next->end = &block_decoder_end;

	if (lzma_check_init(&next->coder->check, next->coder->options.check))
		return LZMA_UNSUPPORTED_CHECK;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_block_decoder(lzma_stream *strm, const lzma_options_block *options)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FLUSH] = true;

	ret = lzma_block_decoder_init(
			&strm->internal->next, strm->allocator, options);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
