///////////////////////////////////////////////////////////////////////////////
//
/// \file       block_header_encoder.c
/// \brief      Encoder Block Header for .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "common.h"
#include "check.h"


extern LZMA_API lzma_ret
lzma_block_header_encode(uint8_t **header, size_t *header_size,
		lzma_allocator *allocator, const lzma_options_block *options)
{
	// Initialize the size of the encoded Block Header (*header)
	// to have enough space for Block Flags, Compressed Size,
	// Uncompressed Size, CRC32, and Padding.
	*header_size = 2 + 2 * LZMA_VLI_BYTES_MAX + 4 + 31;

	// Index for options[] and prop_sizes[]
	size_t i = 0;

	// Calculate the encoded size of the List of Filter Flags.
	while (options->filters[i].id != UINT64_MAX) {
		// Don't allow too many filters.
		if (i == 7)
			return LZMA_HEADER_ERROR;

		const size_t add = lzma_filter_flags_size(
				options->filters + i);
		if (add == 0)
			return LZMA_HEADER_ERROR;

		*header_size += add;
		++i;
	}

	// Allocate memory to hold the encoded form of Block Header
	// excluding Header Padding.
	*header = lzma_alloc(*header_size, allocator);
	if (*header == NULL)
		return LZMA_MEM_ERROR;

	// Write position in header
	size_t header_pos = 0;

	// Write Block Flags with Size of Header Padding set to zero for now.
	assert(i <= 7);
	(*header)[header_pos] = i;

	if (options->has_eopm || options->uncompressed_size == UINT64_MAX)
		(*header)[header_pos] |= 0x08;

	if (options->compressed_size != UINT64_MAX)
		(*header)[header_pos] |= 0x10;

	if (options->uncompressed_size != UINT64_MAX)
		(*header)[header_pos] |= 0x20;

	if (options->is_metadata)
		(*header)[header_pos] |= 0x80;

	++header_pos;
	(*header)[header_pos++] = 0x00;

	// Write Compressed Size and Uncompressed Size.
	if (options->compressed_size != UINT64_MAX) {
		size_t vli_pos = 0;
		if (lzma_vli_encode(options->compressed_size, &vli_pos, 1,
				*header, &header_pos, *header_size))
			goto header_error;
	}

	if (options->uncompressed_size != UINT64_MAX) {
		size_t vli_pos = 0;
		if (lzma_vli_encode(options->uncompressed_size, &vli_pos, 1,
					*header, &header_pos, *header_size)
				!= LZMA_STREAM_END)
			goto header_error;
	}

	// Encode the List of Filter Flags.
	for (i = 0; options->filters[i].id != UINT64_MAX; ++i)
		if (lzma_filter_flags_encode(*header, &header_pos, *header_size,
				options->filters + i) != LZMA_OK)
			goto header_error;

	// Set the real size of the Block Header excluding Header Padding.
	assert(header_pos + 4 <= *header_size);
	*header_size = header_pos;
	if (options->has_crc32)
		*header_size += 4;

	// Calculate the size of the Header Padding, and store it to
	// (*header)[1]. Update *header_size and zero appropriate number
	// of bytes in the end of the *header.
	{
		size_t preferred;
		if (options->filters[0].id == UINT64_MAX) {
			// We have no filters specified, so the actual filter
			// is implicit Copy filter.
			assert(i == 0);
			preferred = 1;
		} else {
			// We have at least one real filter specified. Check
			// the preferred alignment of the output data of
			// the last filter in the chain.
			assert(i >= 1);
			switch (options->filters[i - 1].id) {
			case LZMA_FILTER_POWERPC:
			case LZMA_FILTER_ARM:
			case LZMA_FILTER_SPARC:
				preferred = 4;
				break;

			case LZMA_FILTER_IA64:
				preferred = 16;
				break;

			case LZMA_FILTER_ARMTHUMB:
				preferred = 2;
				break;

			default:
				preferred = 1;
				break;
			}
		}

		const size_t unaligned = *header_size
				+ (size_t)(options->alignment);
		size_t padding = unaligned % preferred;
		if (padding != 0)
			padding = preferred - padding;

		assert(padding <= 31);
		(*header)[1] |= padding;

		memzero(*header + *header_size, padding);
		*header_size += padding;
	}

	// Calculate the CRC32 of the Block Header.
	if (options->has_crc32) {
		assert(header_pos + 4 == *header_size);
		const uint32_t crc = lzma_crc32(*header, header_pos, 0);
		for (size_t i = 0; i < 4; ++i)
			(*header)[header_pos++] = crc >> (i * 8);
	}

	assert(header_pos == *header_size);

	return LZMA_OK;

header_error:
	lzma_free(*header, allocator);
	*header = NULL;
	return LZMA_HEADER_ERROR;
}
