///////////////////////////////////////////////////////////////////////////////
//
/// \file       metadata_encoder.c
/// \brief      Encodes metadata to be stored into Metadata Blocks
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "metadata_encoder.h"


struct lzma_metadata_encoder_s {
	enum {
		SEQ_FLAGS,
		SEQ_HEADER_METADATA_SIZE,
		SEQ_TOTAL_SIZE,
		SEQ_UNCOMPRESSED_SIZE,
		SEQ_INDEX_SIZE,
		SEQ_INDEX_TOTAL,
		SEQ_INDEX_UNCOMPRESSED,
		SEQ_EXTRA_ID,
		SEQ_EXTRA_SIZE,
		SEQ_EXTRA_DATA,
		SEQ_END,
	} sequence;

	/// Position in variable-length integers
	size_t pos;

	/// Local copy of the Metadata structure. Note that we keep
	/// a copy only of the main structure, not Index or Extra Records.
	lzma_metadata metadata;

	/// Number of Records in Index
	size_t index_size;

	/// Index Record currently being processed
	lzma_index *index_current;

	/// True once everything except compression has been done.
	bool end_was_reached;

	/// buffer[buffer_pos] is the first byte that needs to be compressed.
	size_t buffer_pos;

	/// buffer[buffer_size] is the next position where a byte will be
	/// written by process().
	size_t buffer_size;

	/// Temporary buffer to which encoded Metadata is written before
	/// it is compressed.
	uint8_t buffer[LZMA_BUFFER_SIZE];

};


#define write_vli(num) \
do { \
	const lzma_ret ret = lzma_vli_encode(coder->num, &coder->pos, 1, \
			coder->buffer, &coder->buffer_size, \
			LZMA_BUFFER_SIZE); \
	if (ret != LZMA_STREAM_END) \
		return ret; \
	coder->pos = 0; \
} while (0)


static lzma_ret
process(lzma_metadata_encoder *coder)
{
	while (coder->buffer_size < LZMA_BUFFER_SIZE)
	switch (coder->sequence) {
	case SEQ_FLAGS:
		coder->buffer[coder->buffer_size] = 0;

		if (coder->metadata.header_metadata_size != UINT64_MAX)
			coder->buffer[coder->buffer_size] |= 0x01;

		if (coder->metadata.total_size != UINT64_MAX)
			coder->buffer[coder->buffer_size] |= 0x02;

		if (coder->metadata.uncompressed_size != UINT64_MAX)
			coder->buffer[coder->buffer_size] |= 0x04;

		if (coder->metadata.index != NULL)
			coder->buffer[coder->buffer_size] |= 0x08;

		if (coder->metadata.extra != NULL)
			coder->buffer[coder->buffer_size] |= 0x80;

		++coder->buffer_size;
		coder->sequence = SEQ_HEADER_METADATA_SIZE;
		break;

	case SEQ_HEADER_METADATA_SIZE:
		if (coder->metadata.header_metadata_size != UINT64_MAX) {
			write_vli(metadata.header_metadata_size);
			coder->sequence = SEQ_TOTAL_SIZE;
			break;
		}

		coder->sequence = SEQ_TOTAL_SIZE;

	// Fall through

	case SEQ_TOTAL_SIZE:
		if (coder->metadata.total_size != UINT64_MAX) {
			write_vli(metadata.total_size);
			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
			break;
		}

		coder->sequence = SEQ_UNCOMPRESSED_SIZE;

	// Fall through

	case SEQ_UNCOMPRESSED_SIZE:
		if (coder->metadata.uncompressed_size != UINT64_MAX) {
			write_vli(metadata.uncompressed_size);
			coder->sequence = SEQ_INDEX_SIZE;
			break;
		}

		coder->sequence = SEQ_INDEX_SIZE;

	// Fall through

	case SEQ_INDEX_SIZE:
		if (coder->index_size == 0) {
			if (coder->metadata.extra == NULL) {
				coder->sequence = SEQ_END;
				return LZMA_STREAM_END;
			}

			coder->sequence = SEQ_EXTRA_ID;
			break;
		}

		write_vli(index_size);
		coder->index_current = coder->metadata.index;
		coder->sequence = SEQ_INDEX_TOTAL;
		break;

	case SEQ_INDEX_TOTAL:
		write_vli(index_current->total_size);

		coder->index_current = coder->index_current->next;
		if (coder->index_current == NULL) {
			coder->index_current = coder->metadata.index;
			coder->sequence = SEQ_INDEX_UNCOMPRESSED;
		}

		break;

	case SEQ_INDEX_UNCOMPRESSED:
		write_vli(index_current->uncompressed_size);

		coder->index_current = coder->index_current->next;

		if (coder->index_current != NULL)
			break;

		if (coder->metadata.extra != NULL) {
			coder->sequence = SEQ_EXTRA_ID;
			break;
		}

		coder->sequence = SEQ_END;
		return LZMA_STREAM_END;

	case SEQ_EXTRA_ID: {
		const lzma_ret ret = lzma_vli_encode(
				coder->metadata.extra->id, &coder->pos, 1,
				coder->buffer, &coder->buffer_size,
				LZMA_BUFFER_SIZE);
		switch (ret) {
		case LZMA_OK:
			break;

		case LZMA_STREAM_END:
			coder->pos = 0;

			// Handle the special ID 0.
			if (coder->metadata.extra->id == 0) {
				coder->metadata.extra
						= coder->metadata.extra->next;
				if (coder->metadata.extra == NULL) {
					coder->sequence = SEQ_END;
					return LZMA_STREAM_END;
				}

				coder->sequence = SEQ_EXTRA_ID;

			} else {
				coder->sequence = SEQ_EXTRA_SIZE;
			}

			break;

		default:
			return ret;
		}

		break;
	}

	case SEQ_EXTRA_SIZE:
		if (coder->metadata.extra->size >= (uint64_t)(SIZE_MAX))
			return LZMA_HEADER_ERROR;

		write_vli(metadata.extra->size);
		coder->sequence = SEQ_EXTRA_DATA;
		break;

	case SEQ_EXTRA_DATA:
		bufcpy(coder->metadata.extra->data, &coder->pos,
				coder->metadata.extra->size,
				coder->buffer, &coder->buffer_size,
				LZMA_BUFFER_SIZE);

		if ((size_t)(coder->metadata.extra->size) == coder->pos) {
			coder->metadata.extra = coder->metadata.extra->next;
			if (coder->metadata.extra == NULL) {
				coder->sequence = SEQ_END;
				return LZMA_STREAM_END;
			}

			coder->pos = 0;
			coder->sequence = SEQ_EXTRA_ID;
		}

		break;

	case SEQ_END:
		// Everything is encoded. Let the compression code finish
		// its work now.
		return LZMA_STREAM_END;
	}

	return LZMA_OK;
}


extern lzma_ret
lzma_metadata_encode_internal(lzma_metadata_encoder *coder,
		lzma_next_coder *next, lzma_allocator *allocator,
		uint8_t *out, size_t *out_pos, size_t out_size)
{
	while (!coder->end_was_reached) {
		// Flush coder->buffer if it isn't empty.
		if (coder->buffer_size > 0) {
			const lzma_ret ret = next->code(
					next->coder, allocator,
					coder->buffer, &coder->buffer_pos,
					coder->buffer_size,
					out, out_pos, out_size, LZMA_RUN);
			if (coder->buffer_pos < coder->buffer_size
					|| ret != LZMA_OK)
				return ret;

			coder->buffer_pos = 0;
			coder->buffer_size = 0;
		}

		const lzma_ret ret = process(coder);

		switch (ret) {
		case LZMA_OK:
			break;

		case LZMA_STREAM_END:
			coder->end_was_reached = true;
			break;

		default:
			return ret;
		}
	}

	// Finish
	return next->code(next->coder, allocator, coder->buffer,
			&coder->buffer_pos, coder->buffer_size,
			out, out_pos, out_size, LZMA_FINISH);
}


extern LZMA_API lzma_ret
lzma_metadata_encoder_init(lzma_metadata_encoder **coder,
		lzma_allocator *allocator, const lzma_metadata *metadata)
{
	*coder = lzma_alloc(sizeof(lzma_metadata_encoder), allocator);
	if (*coder == NULL)
		return LZMA_MEM_ERROR;

	(*coder)->sequence = SEQ_FLAGS;
	(*coder)->pos = 0;
	(*coder)->metadata = *metadata;

	(*coder)->index_size = 0;
	for (const lzma_index *i = (*coder)->metadata.index;
			i != NULL; i = i->next)
		++(*coder)->index_size;

	(*coder)->end_was_reached = false;
	(*coder)->buffer_pos = 0;
	(*coder)->buffer_size = 0;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_metadata_encode(lzma_metadata_encoder *coder, lzma_stream *strm)
{
	size_t out_pos = 0;
	const lzma_ret ret = lzma_metadata_encode_internal(coder,
			&strm->internal->next, strm->allocator,
			strm->next_out, &out_pos, strm->avail_out);

	strm->next_out += out_pos;
	strm->avail_out -= out_pos;

	return ret;
}


extern LZMA_API void
lzma_metadata_encoder_end(
		lzma_metadata_encoder **coder, lzma_allocator *allocator)
{
	lzma_free(*coder, allocator);
	*coder = NULL;
	return;
}


extern LZMA_API uint64_t
lzma_metadata_size(const lzma_metadata *metadata)
{
	uint64_t size = 1; // Metadata Flags

	// Validate header_metadata_size, total_size, and uncompressed_size.
	if (!lzma_vli_is_valid(metadata->header_metadata_size)
			|| !lzma_vli_is_valid(metadata->total_size)
			|| !lzma_vli_is_valid(metadata->uncompressed_size))
		return 0;

	// Add the sizes of these three fields.
	if (metadata->header_metadata_size != UINT64_MAX)
		size += lzma_vli_size(metadata->header_metadata_size);

	if (metadata->total_size != UINT64_MAX)
		size += lzma_vli_size(metadata->total_size);

	if (metadata->uncompressed_size != UINT64_MAX)
		size += lzma_vli_size(metadata->uncompressed_size);

	// Index
	{
		const lzma_index *i = metadata->index;
		uint64_t index_size = 0;

		while (i != NULL) {
			// Validate the sizes.
			if (!lzma_vli_is_valid(i->uncompressed_size)
					|| !lzma_vli_is_valid(i->total_size))
				return 0;

			// Add the sizes of these two fields.
			if (i->total_size != UINT64_MAX)
				size += lzma_vli_size(i->total_size);

			if (i->uncompressed_size != UINT64_MAX)
				size += lzma_vli_size(i->uncompressed_size);

			i = i->next;
			++index_size;
		}

		if (index_size > 0) {
			const size_t add = lzma_vli_size(index_size);
			if (add == 0)
				return 0;

			size += add;
		}
	}

	// Extra
	{
		const lzma_extra *e = metadata->extra;
		while (e != NULL) {
			// Validate the numbers.
			if (e->id > LZMA_VLI_VALUE_MAX
					|| e->size >= (uint64_t)(SIZE_MAX))
				return 0;

			// Add the sizes.
			size += lzma_vli_size(e->id);
			if (e->id != 0) {
				size += lzma_vli_size(e->size);
				size += e->size;
			}

			e = e->next;
		}
	}

	return size;
}
