///////////////////////////////////////////////////////////////////////////////
//
/// \file       raw_common.c
/// \brief      Stuff shared between raw encoder and raw decoder
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "raw_common.h"


/// \brief      Prepares the filter chain
///
/// Prepares the filter chain by setting uncompressed sizes for each filter,
/// and adding implicit Subblock filter when needed.
///
/// \return     true if error occurred, false on success.
///
static bool
prepare(uint64_t *id, uint64_t *uncompressed_size, bool implicit)
{
	bool needs_end_of_input = false;

	switch (id[0]) {
	case LZMA_FILTER_COPY:
	case LZMA_FILTER_X86:
	case LZMA_FILTER_POWERPC:
	case LZMA_FILTER_IA64:
	case LZMA_FILTER_ARM:
	case LZMA_FILTER_ARMTHUMB:
	case LZMA_FILTER_SPARC:
		uncompressed_size[1] = uncompressed_size[0];
		needs_end_of_input = true;
		break;

	case LZMA_FILTER_SPARSE:
		// - Uncompressed size must be known.
		// - Must be the last filter.
		if (uncompressed_size[0] == UINT64_MAX || id[1] != UINT64_MAX)
			return true;

		// No encoded output
		uncompressed_size[1] = 0;
		break;

	case LZMA_FILTER_SUBBLOCK:
	case LZMA_FILTER_LZMA:
		// These change the size of the data unpredictably.
		uncompressed_size[1] = UINT64_MAX;
		break;

	case LZMA_FILTER_SUBBLOCK_HELPER:
		uncompressed_size[1] = uncompressed_size[0];
		break;

	default:
		// Unknown filter.
		return true;
	}

	// Is this the last filter in the chain?
	if (id[1] == UINT64_MAX) {
		if (!needs_end_of_input || !implicit
				|| uncompressed_size[0] != UINT64_MAX)
			return false;

		// Add implicit Subblock filter.
		id[1] = LZMA_FILTER_SUBBLOCK;
		uncompressed_size[1] = UINT64_MAX;
		id[2] = UINT64_MAX;
	}

	return prepare(id + 1, uncompressed_size + 1, implicit);
}


extern lzma_ret
lzma_raw_coder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_options_filter *options, uint64_t uncompressed_size,
		lzma_init_function (*get_function)(uint64_t id),
		bool implicit, bool is_encoder)
{
	// Count the number of filters in the chain.
	size_t count = 0;
	while (options[count].id != UINT64_MAX)
		++count;

	// Allocate enough space from the stack for IDs and uncompressed
	// sizes. We need two extra: possible implicit Subblock and end
	// of array indicator.
	uint64_t ids[count + 2];
	uint64_t uncompressed_sizes[count + 2];

	uncompressed_sizes[0] = uncompressed_size;

	if (count == 0) {
		if (!implicit)
			return LZMA_HEADER_ERROR;

		count = 1;

		// Special case: no filters were specified, so an implicit
		// Copy or Subblock filter is used.
		if (uncompressed_size == UINT64_MAX)
			ids[0] = LZMA_FILTER_SUBBLOCK;
		else
			ids[0] = LZMA_FILTER_COPY;

		ids[1] = UINT64_MAX;

	} else {
		// Prepare the ids[] and uncompressed_sizes[].
		for (size_t i = 0; i < count; ++i)
			ids[i] = options[i].id;

		ids[count] = UINT64_MAX;

		if (prepare(ids, uncompressed_sizes, implicit))
			return LZMA_HEADER_ERROR;

		// Check if implicit Subblock filter was added.
		if (ids[count] != UINT64_MAX) {
			assert(ids[count] == LZMA_FILTER_SUBBLOCK);
			++count;
		}
	}

	// Allocate space from the stack for filter information.
	lzma_filter_info filters[count + 1];

	// Set the filter functions, and copy uncompressed sizes and options.
	if (is_encoder) {
		for (size_t i = 0; i < count; ++i) {
			// The order of the filters is reversed in the
			// encoder. It allows more efficient handling
			// of the uncompressed data.
			const size_t j = count - i - 1;

			filters[j].init = get_function(ids[i]);
			if (filters[j].init == NULL) {
				lzma_next_coder_end(next, allocator);
				return LZMA_HEADER_ERROR;
			}

			filters[j].options = options[i].options;
			filters[j].uncompressed_size = uncompressed_sizes[i];
		}

	} else {
		for (size_t i = 0; i < count; ++i) {
			filters[i].init = get_function(ids[i]);
			if (filters[i].init == NULL) {
				lzma_next_coder_end(next, allocator);
				return LZMA_HEADER_ERROR;
			}

			filters[i].options = options[i].options;
			filters[i].uncompressed_size = uncompressed_sizes[i];
		}
	}

	// Terminate the array.
	filters[count].init = NULL;

	// Initialize the filters.
	return lzma_next_coder_init(next, allocator, filters);
}
