///////////////////////////////////////////////////////////////////////////////
//
/// \file       stream_decoder.c
/// \brief      Decodes .lzma Streams
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "stream_common.h"
#include "check.h"
#include "stream_flags_decoder.h"
#include "block_decoder.h"
#include "metadata_decoder.h"


struct lzma_coder_s {
	lzma_next_coder next;

	enum {
		SEQ_STREAM_HEADER_CODE,
		SEQ_HEADER_METADATA_INIT,
		SEQ_HEADER_METADATA_CODE,
		SEQ_PADDING,
		SEQ_DATA_INIT_1,
		SEQ_DATA_INIT_2,
		SEQ_DATA_CODE,
		SEQ_FOOTER_METADATA_INIT,
		SEQ_FOOTER_METADATA_CODE,
		SEQ_STREAM_FOOTER_INIT,
		SEQ_STREAM_FOOTER_CODE,
	} sequence;

	/// Position in variable-length integers and in some other things.
	size_t pos;

	/// Uncompressed Size of the Block currently being processed.
	uint64_t uncompressed_size;

	/// Total Size of the Block currently being processed.
	uint64_t total_size;

	/// Real size of the Header Metadata Block
	uint64_t header_metadata_size;

	/// Real size of the Footer Metadata Block
	uint64_t footer_metadata_size;

	/// Stream Flags from Stream Header
	lzma_stream_flags header_flags;

	/// Stream Flags from Stream Footer
	lzma_stream_flags footer_flags;

	/// Decoder for Stream Header and Stream Footer. This takes very
	/// little memory and the same data structure can be used for
	/// both Header and Footer, so it's a good idea to have a separate
	/// lzma_next_coder structure for it.
	lzma_next_coder flags_decoder;

	/// Decoder for Metadata Blocks
	lzma_metadata_decoder *metadata_decoder;

	/// Presence of the Uncompressed Size field in Block Header of
	/// Single-Block Stream. This is needed to know if Uncompressed Size
	/// is present in Stream Footer.
	bool block_header_has_uncompressed_size;

	/// Presence of the Index field in Header Metadata Block. This is
	/// used to check if we need to add Records to Index or compare to
	/// existing values. FIXME desc
	bool has_index_in_header_metadata;

	/// Index Record of the currently processed Block.
	lzma_index *index_current;

	lzma_metadata header_metadata;
	lzma_metadata footer_metadata;
	lzma_extra **header_extra;
	lzma_extra **footer_extra;
};


static lzma_ret
stream_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size && (*in_pos < in_size
			// FIXME Since there is always a Footer, we can
			// simplify this by assuming that there must always
			// be at least one byte of input.
			|| coder->sequence == SEQ_DATA_CODE
			|| coder->sequence == SEQ_HEADER_METADATA_CODE
			|| coder->sequence == SEQ_FOOTER_METADATA_CODE))
	switch (coder->sequence) {
	case SEQ_STREAM_HEADER_CODE: {
		size_t dummy = 0;
		const lzma_ret ret = coder->flags_decoder.code(
				coder->flags_decoder.coder,
				allocator, in, in_pos, in_size,
				NULL, &dummy, 0, LZMA_RUN);
		if (ret != LZMA_STREAM_END)
			return ret;

		if (coder->header_flags.is_multi)
			coder->sequence = SEQ_HEADER_METADATA_INIT;
		else
			coder->sequence = SEQ_DATA_INIT_1;

		// Detect if the Check type is supported and give appropriate
		// warning if it isn't. We don't warn every time a new Block
		// is started.
		lzma_check tmp;
		if (lzma_check_init(&tmp, coder->header_flags.check))
			return LZMA_UNSUPPORTED_CHECK;

		break;
	}

	case SEQ_HEADER_METADATA_INIT:
		// Header Metadata Block is optional.
		if (!(in[*in_pos] & 0x80)) {
			coder->sequence = SEQ_DATA_INIT_1;
			break;
		}

	// Fall through

	case SEQ_FOOTER_METADATA_INIT: {
		lzma_options_block options = {
			.check = coder->header_flags.check,
			.has_crc32 = coder->header_flags.has_crc32,
			.uncompressed_size = UINT64_MAX,
			.compressed_size = UINT64_MAX,
		};
		lzma_ret ret = lzma_block_decoder_init(&coder->next,
				allocator, &options);
		if (ret != LZMA_OK && ret != LZMA_UNSUPPORTED_CHECK)
			return ret;

		// This needs to be initialized after Block decoder, because
		// Metadata decoder doesn't automatically free memory if
		// initialized multiple times.
		ret = lzma_metadata_decoder_init(
				&coder->metadata_decoder, allocator,
				coder->sequence == SEQ_HEADER_METADATA_INIT
					? &coder->header_metadata
					: &coder->footer_metadata,
				(coder->sequence == SEQ_HEADER_METADATA_INIT
						? coder->header_extra
						: coder->footer_extra)
					!= NULL);
		if (ret != LZMA_OK)
			return ret;

		++coder->sequence;
		break;
	}

	case SEQ_HEADER_METADATA_CODE:
	case SEQ_FOOTER_METADATA_CODE: {
		const size_t in_start = *in_pos;

		const lzma_ret ret = lzma_metadata_decode_real(
				coder->metadata_decoder, &coder->next,
				allocator, in, in_pos, in_size);

		const size_t in_used = *in_pos - in_start;
		if (coder->sequence == SEQ_HEADER_METADATA_CODE)
			coder->header_metadata_size += in_used;
		else
			coder->footer_metadata_size += in_used;

		if (ret != LZMA_STREAM_END)
			return ret;

		lzma_metadata_decoder_end(&coder->metadata_decoder, allocator);

		if (coder->sequence == SEQ_HEADER_METADATA_CODE) {
			if (coder->header_extra != NULL) {
				*coder->header_extra
						= coder->header_metadata.extra;
				coder->header_metadata.extra = NULL;
			}

			assert(coder->header_metadata.extra == NULL);

			// The value of Header Metadata Size is ignored in
			// Header Metadata Block. We use it to hold the
			// real size of Header Metadata Block, which will
			// be compared to the value stored in Footer Metadata.
			coder->header_metadata.header_metadata_size
					= coder->total_size;

			coder->sequence = SEQ_PADDING;

		} else {
			if (coder->footer_extra != NULL) {
				*coder->footer_extra
						= coder->footer_metadata.extra;
				coder->footer_metadata.extra = NULL;
			}

			assert(coder->footer_metadata.extra == NULL);

			if (!lzma_metadata_index_equal(
					coder->header_metadata.index,
					coder->footer_metadata.index))
				return LZMA_DATA_ERROR;

			coder->sequence = SEQ_STREAM_FOOTER_INIT;
		}

		break;
	}

	case SEQ_DATA_INIT_1: {
		// Detect Metadata Block.
		if (in[*in_pos] & 0x80) {
			if (coder->header_flags.is_multi) {
				// There must be at least one Data Block in
				// Multi-Block Stream.
				if (coder->index_current == NULL)
					return LZMA_DATA_ERROR;

				// If Index was stored in Header Metadata
				// Block, we must now have processed all
				// Data Blocks.
				if (coder->index_current->next != NULL)
					return LZMA_DATA_ERROR;

			} else {
				// The Block of Single-Block Stream
				// must not be a Metadata Block.
				return LZMA_DATA_ERROR;
			}

			coder->sequence = SEQ_FOOTER_METADATA_INIT;
			break;
		}

		if (coder->header_flags.is_multi) {
			if (coder->has_index_in_header_metadata) {
				// Forward to the next Index entry. If next
				// entry doesn't exist, Index doesn't match
				// the Stream contents, and error is returned.
				if (coder->index_current == NULL)
					coder->index_current
						= coder->header_metadata.index;
				else if (coder->index_current->next == NULL)
					return LZMA_DATA_ERROR;
				else
					coder->index_current
						= coder->index_current->next;
			} else {
				// Allocate and initialize a new Index Record.
				lzma_index *i = lzma_alloc(
						sizeof(lzma_index), allocator);
				if (i == NULL)
					return LZMA_MEM_ERROR;

				i->next = NULL;
				i->total_size = 0;
				i->uncompressed_size = 0;

				if (coder->header_metadata.index == NULL)
					coder->header_metadata.index = i;
				else
					coder->index_current->next = i;

				coder->index_current = i;
			}
		} else {
			// Check presence of the Uncompressed Size field. It is
			// needed with Single-Block Streams to know if
			// Uncompressed Size field is present in Stream Footer.
			coder->block_header_has_uncompressed_size
					= (in[*in_pos] & 0x20) != 0;
		}

		coder->sequence = SEQ_DATA_INIT_2;
	}

	// Fall through

	case SEQ_DATA_INIT_2: {
		// Initialize Block decoder.
		lzma_options_block options = {
			.check = coder->header_flags.check,
			.has_crc32 = coder->header_flags.has_crc32,
			.compressed_size = UINT64_MAX,
		};

		if (coder->has_index_in_header_metadata)
			options.uncompressed_size = coder->index_current
					->uncompressed_size;
		else
			options.uncompressed_size = UINT64_MAX;

		const lzma_ret ret = lzma_block_decoder_init(&coder->next,
				allocator, &options);
		if (ret != LZMA_OK && ret != LZMA_UNSUPPORTED_CHECK)
			return ret;

		coder->sequence = SEQ_DATA_CODE;
	}

	// Fall through

	case SEQ_DATA_CODE: {
		// Store old positions. They are needed to update
		// coder->uncompressed_size and coder->total_size.
		const size_t in_start = *in_pos;
		const size_t out_start = *out_pos;

		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		const size_t in_used = *in_pos - in_start;
		const size_t out_used = *out_pos - out_start;

		coder->total_size += in_used;
		coder->uncompressed_size += out_used;

		if (coder->index_current != NULL) {
			coder->index_current->total_size += in_used;
			coder->index_current->uncompressed_size += out_used;
		}

		if (ret != LZMA_STREAM_END)
			return ret;

		// In a Multi-Block Stream, Data Block is followed
		// with Padding. In Single-Block Stream the Stream
		// Footer follows immediatelly.
		if (coder->header_flags.is_multi)
			coder->sequence = SEQ_PADDING;
		else
			coder->sequence = SEQ_STREAM_FOOTER_INIT;

		break;
	}

	case SEQ_PADDING:
		if (in[*in_pos] == 0x00) {
			++coder->total_size;

			if (coder->index_current != NULL)
				++coder->index_current->total_size;
			else
				++coder->header_metadata_size;

			++*in_pos;
			break;
		}

		coder->sequence = SEQ_DATA_INIT_1;
		break;

	case SEQ_STREAM_FOOTER_INIT: {
		const lzma_ret ret = lzma_stream_footer_decoder_init(
				&coder->flags_decoder, allocator,
				&coder->footer_flags,
				!coder->header_flags.is_multi
				&& !coder->block_header_has_uncompressed_size);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_STREAM_FOOTER_CODE;
	}

	// Fall through

	case SEQ_STREAM_FOOTER_CODE: {
		const lzma_ret ret = coder->flags_decoder.code(
				coder->flags_decoder.coder, allocator,
				in, in_pos, in_size, NULL, NULL, 0, LZMA_RUN);
		if (ret != LZMA_STREAM_END)
			return ret;

		// With Single-Block Streams that don't have Uncompressed Size
		// in Stream Header, we must verify that Uncompressed Size
		// in Block Footer matches the real uncompressed size.
		if (!coder->header_flags.is_multi
				&& !coder->block_header_has_uncompressed_size
				&& coder->footer_flags.uncompressed_size
					!= coder->uncompressed_size)
			return LZMA_DATA_ERROR;

		// Verify the Backward Size field.
		uint64_t backward_size;
		if (coder->header_flags.is_multi) {
			backward_size = coder->footer_metadata_size;
		} else {
			backward_size = coder->total_size;

			// If Uncompressed Size is present in Stream Footer,
			// we must take the size of that field into account.
			if (!coder->block_header_has_uncompressed_size)
				backward_size += lzma_vli_size(
						coder->uncompressed_size);
		}
		if (backward_size != coder->footer_flags.backward_size)
			return LZMA_DATA_ERROR;

		// Compare that Stream Flags in Stream Footer match
		// the same field in Stream Header.
		if (coder->header_flags.check != coder->footer_flags.check
				|| coder->header_flags.has_crc32
					!= coder->footer_flags.has_crc32
				|| coder->header_flags.is_multi
					!= coder->footer_flags.is_multi)
			return LZMA_DATA_ERROR;

		return LZMA_STREAM_END;
	}

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
free_index(lzma_index *index, lzma_allocator *allocator)
{
	while (index != NULL) {
		lzma_index *tmp = index;
		index = index->next;
		lzma_free(tmp, allocator);
	}

	return;
}


static void
free_substructures(lzma_coder *coder, lzma_allocator *allocator)
{
	free_index(coder->header_metadata.index, allocator);
	free_index(coder->footer_metadata.index, allocator);
	coder->header_metadata.index = NULL;
	coder->header_metadata.extra = NULL;
	coder->footer_metadata.index = NULL;
	coder->footer_metadata.extra = NULL;
	coder->index_current = NULL;

	lzma_metadata_decoder_end(&coder->metadata_decoder, allocator);

	return;
}


static void
stream_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	free_substructures(coder, allocator);
	lzma_next_coder_end(&coder->flags_decoder, allocator);
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_stream_decoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		lzma_extra **header, lzma_extra **footer)
{
	if (next->init != &lzma_stream_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		// Mark that we haven't allocated memory for any substructure.
		next->coder->next = LZMA_NEXT_CODER_INIT;
		next->coder->flags_decoder = LZMA_NEXT_CODER_INIT;
		next->coder->metadata_decoder = NULL;
		next->coder->header_metadata.index = NULL;
		next->coder->header_metadata.extra = NULL;
		next->coder->footer_metadata.index = NULL;
		next->coder->footer_metadata.extra = NULL;
	} else {
		// Free memory of existing substructures.
		free_substructures(next->coder, allocator);
	}

	// Initialize Stream Header decoder.
	{
		const lzma_ret ret = lzma_stream_header_decoder_init(
				&next->coder->flags_decoder, allocator,
				&next->coder->header_flags);
		if (ret != LZMA_OK) {
			stream_decoder_end(next->coder, allocator);
			return ret;
		}
	}

	// Reset some variables.
	next->coder->sequence = SEQ_STREAM_HEADER_CODE;
	next->coder->pos = 0;
	next->coder->uncompressed_size = 0;
	next->coder->total_size = 0;
	next->coder->header_metadata_size = 0;
	next->coder->footer_metadata_size = 0;
	next->coder->block_header_has_uncompressed_size = false;
	next->coder->has_index_in_header_metadata = false;
	next->coder->index_current = NULL;
	next->coder->header_extra = header;
	next->coder->footer_extra = footer;

	// Set the function pointers.
	next->init = &lzma_stream_decoder_init;
	next->code = &stream_decode;
	next->end = &stream_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_decoder(lzma_stream *strm,
		lzma_extra **header, lzma_extra **footer)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FLUSH] = true;

	ret = lzma_stream_decoder_init(&strm->internal->next, strm->allocator,
			header, footer);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
