///////////////////////////////////////////////////////////////////////////////
//
/// \file       stream_encoder_multi.c
/// \brief      Encodes Multi-Block .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "stream_common.h"
#include "block_encoder.h"
#include "metadata_encoder.h"


struct lzma_coder_s {
	lzma_next_coder next;

	lzma_metadata_encoder *metadata_encoder;

	const lzma_options_stream *options;

	enum {
		SEQ_STREAM_HEADER_COPY,
		SEQ_HEADER_METADATA_INIT,
		SEQ_HEADER_METADATA_CODE,
		SEQ_DATA_INIT_1,
		SEQ_DATA_INIT_2,
		SEQ_DATA_CODE,
		SEQ_FOOTER_METADATA_INIT,
		SEQ_FOOTER_METADATA_CODE,
		SEQ_STREAM_FOOTER_INIT,
		SEQ_STREAM_FOOTER_COPY,
	} sequence;

	/// Local copy of options->check to prevent application from changing
	/// this option in the midd le of encoding.
	lzma_check_type check;

	/// Like `check' above.
	bool has_crc32;

	/// Variable to track current output position relative to the
	/// beginning of the Stream. This way alignment of Compressed Data
	/// in Blocks can be kept correct.
	uint32_t alignment;

	/// Size of the Header Metadata Block. This is needed to when
	/// writing Footer Metadata Block.
	uint64_t header_metadata_size;

	/// Backward Size is the Total Size of Footer Metadata Block.
	uint64_t backward_size;

	/// First Index Record
	lzma_index *index_head;

	/// Last Index Record
	lzma_index *index_tail;

	/// Number of Records in Index so far
	size_t index_size;

	/// Strema Header or Stream Footer in encoded form
	uint8_t *header;
	size_t header_pos;
	size_t header_size;
};


static lzma_ret
stream_encode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size)
	switch (coder->sequence) {
	case SEQ_STREAM_FOOTER_INIT: {
		lzma_stream_flags flags = {
			.check = coder->check,
			.has_crc32 = coder->has_crc32,
			.is_multi = true,
			.backward_size = coder->backward_size,
		};
		const lzma_ret ret = lzma_stream_footer_encode(
				&coder->header, &coder->header_size,
				allocator, &flags);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_STREAM_FOOTER_COPY;
	}

	// Fall through

	case SEQ_STREAM_HEADER_COPY:
	case SEQ_STREAM_FOOTER_COPY:
		bufcpy(coder->header, &coder->header_pos, coder->header_size,
				out, out_pos, out_size);

		if (coder->header_pos == coder->header_size) {
			lzma_free(coder->header, allocator);
			coder->header = NULL;
			coder->header_pos = 0;

			if (coder->sequence == SEQ_STREAM_FOOTER_COPY)
				return LZMA_STREAM_END;

			if (coder->options->header == NULL)
				coder->sequence = SEQ_DATA_INIT_1;
			else
				coder->sequence = SEQ_HEADER_METADATA_INIT;
		}

		break;

	case SEQ_HEADER_METADATA_INIT:
	case SEQ_FOOTER_METADATA_INIT: {
		lzma_extra *extra = coder->sequence == SEQ_HEADER_METADATA_INIT
				? coder->options->header
				: coder->options->footer;

		lzma_metadata metadata = {
			.header_metadata_size = coder->header_metadata_size,
			.total_size = UINT64_MAX,
			.uncompressed_size = UINT64_MAX,
			.index = coder->index_head,
			.extra = extra,
		};

		const uint64_t metadata_size = lzma_metadata_size(&metadata);
		if (metadata_size == 0)
			return LZMA_STREAM_ERROR;

		lzma_options_block options = {
			.check = coder->check,
			.has_crc32 = coder->has_crc32,
			.uncompressed_size = metadata_size,
			.compressed_size = UINT64_MAX,
			.has_eopm = false,
			.is_metadata = true,
			.alignment = coder->alignment,
		};
		memcpy(options.filters, coder->options->metadata_filters,
				sizeof(coder->options->filters));

		lzma_ret ret = lzma_block_encoder_init(&coder->next,
				allocator, &options);
		if (ret != LZMA_OK)
			return ret;

		// NOTE: This must be done after initialization of Block
		// encoder, because Metadata encoder doesn't check if
		// metadata_encoder is already allocated, and we want things
		// to be restartable in case of an error.
		ret = lzma_metadata_encoder_init(&coder->metadata_encoder,
				allocator, &metadata);
		if (ret != LZMA_OK)
			return ret;

		if (coder->sequence == SEQ_HEADER_METADATA_INIT)
			coder->header_metadata_size = 0;

		++coder->sequence;
	}

	// Fall through

	case SEQ_HEADER_METADATA_CODE:
	case SEQ_FOOTER_METADATA_CODE: {
		const size_t out_start = *out_pos;

		const lzma_ret ret = lzma_metadata_encode_internal(
				coder->metadata_encoder,
				&coder->next, allocator,
				out, out_pos, out_size);

		const size_t out_used = *out_pos - out_start;
		coder->alignment += out_used;
		if (coder->sequence == SEQ_HEADER_METADATA_CODE)
			coder->header_metadata_size += out_used;
		else
			coder->backward_size += out_used;

		if (ret != LZMA_STREAM_END)
			return ret;

		lzma_metadata_encoder_end(&coder->metadata_encoder, allocator);

		++coder->sequence;
		break;
	}

	case SEQ_DATA_INIT_1: {
		// Don't create a useless empty Block if LZMA_FINISH is
		// used immediatelly after LZMA_FLUSH has finished, except
		// when it would be the only Block.
		if (action == LZMA_FINISH && *in_pos == in_size
				&& coder->index_head != NULL) {
			coder->sequence = SEQ_FOOTER_METADATA_INIT;
			break;
		}

		// Check that we don't go over the limits.
		if (coder->index_size >= lzma_index_size_max)
			return LZMA_LIMIT_REACHED;

		// Allocate and initialize new Index Record. This is kept
		// separate from Block encoder to keep the code restartable
		// in case of allocation fails.
		lzma_index *i = lzma_alloc(sizeof(lzma_index), allocator);
		if (i == NULL)
			return LZMA_MEM_ERROR;

		if (coder->index_head == NULL)
			coder->index_head = i;
		else
			coder->index_tail->next = i;

		coder->index_tail = i;

		i->next = NULL;
		i->total_size = 0;
		i->uncompressed_size = 0;

		++coder->index_size;
		coder->sequence = SEQ_DATA_INIT_2;
	}

	// Fall through

	case SEQ_DATA_INIT_2: {
		// Initialize the encoder chain.
		lzma_options_block options = {
			.check = coder->check,
			.has_crc32 = coder->has_crc32,
			// We don't know the sizes, because the Stream may
			// consist of multiple Data Blocks.
			.uncompressed_size = UINT64_MAX,
			.compressed_size = UINT64_MAX,
			.has_eopm = false,
			.is_metadata = false,
			.alignment = coder->alignment,
		};
		memcpy(options.filters, coder->options->filters,
				sizeof(options.filters));

		const lzma_ret ret = lzma_block_encoder_init(&coder->next,
				allocator, &options);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_DATA_CODE;
	}

	// Fall through

	case SEQ_DATA_CODE: {
		static const lzma_action convert[3] = {
			LZMA_RUN,
			LZMA_FINISH,
			LZMA_FINISH,
		};

		const size_t in_start = *in_pos;
		const size_t out_start = *out_pos;

		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, convert[action]);

		const size_t in_new = *in_pos - in_start;
		const size_t out_new = *out_pos - out_start;

		coder->index_tail->uncompressed_size += in_new;
		coder->index_tail->total_size += out_new;

		coder->alignment += out_new;

		if (ret != LZMA_STREAM_END)
			return ret;

		assert(action == LZMA_FLUSH || action == LZMA_FINISH);

		if (action == LZMA_FLUSH) {
			coder->sequence = SEQ_DATA_INIT_1;
			assert(*in_pos == in_size);
			return LZMA_STREAM_END;
		}

		coder->sequence = SEQ_FOOTER_METADATA_INIT;
		break;
	}

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
free_substructures(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_metadata_encoder_end(&coder->metadata_encoder, allocator);

	lzma_index *i = coder->index_head;
	while (i != NULL) {
		lzma_index *tmp = i;
		i = i->next;
		lzma_free(tmp, allocator);
	}

	lzma_free(coder->header, allocator);

	return;
}


static void
stream_encoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	free_substructures(coder, allocator);
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_stream_encoder_multi_init(lzma_next_coder *next,
		lzma_allocator *allocator, const lzma_options_stream *options)
{
	if (next->init != &lzma_stream_encoder_multi_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		next->coder->next = LZMA_NEXT_CODER_INIT;
	} else {
		free_substructures(next->coder, allocator);
	}

	next->coder->metadata_encoder = NULL;
	next->coder->index_head = NULL;
	next->coder->index_tail = NULL;
	next->coder->header = NULL;

	next->coder->options = options;
	next->coder->sequence = SEQ_STREAM_HEADER_COPY;
	next->coder->check = options->check;
	next->coder->has_crc32 = options->has_crc32;
	next->coder->header_metadata_size = UINT64_MAX;
	next->coder->backward_size = 0;
	next->coder->index_size = 0;
	next->coder->header_pos = 0;
	next->coder->header_size = 0;

	// Encode Stream Flags
	{
		lzma_stream_flags flags = {
			.check = next->coder->check,
			.has_crc32 = next->coder->has_crc32,
			.is_multi = true,
		};
		const lzma_ret ret = lzma_stream_header_encode(
				&next->coder->header,
				&next->coder->header_size, allocator, &flags);
		if (ret != LZMA_OK) {
			stream_encoder_end(next->coder, allocator);
			return ret;
		}

		next->coder->header_pos = 0;
	}

	next->coder->alignment = options->alignment + next->coder->header_size;

	next->init = &lzma_stream_encoder_multi_init;
	next->code = &stream_encode;
	next->end = &stream_encoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_encoder_multi(
		lzma_stream *strm, const lzma_options_stream *options)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FLUSH] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	ret = lzma_stream_encoder_multi_init(&strm->internal->next,
			strm->allocator, options);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}
