///////////////////////////////////////////////////////////////////////////////
//
/// \file       stream_flags_encoder.c
/// \brief      Encodes Stream Header and Footer for .lzma files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "stream_common.h"


static void
stream_flags_encode(uint8_t *out, const lzma_stream_flags *options)
{
	// Check type
	*out = options->check;

	// Usage of CRC32 in Block Headers
	if (options->has_crc32)
		*out |= 0x08;

	// Single- or Multi-Block
	if (options->is_multi)
		*out |= 0x10;

	return;
}


extern LZMA_API lzma_ret
lzma_stream_header_encode(uint8_t **header, size_t *header_size,
		lzma_allocator *allocator, const lzma_stream_flags *options)
{
	*header_size = sizeof(lzma_header_magic) + 1 + 4;
	*header = lzma_alloc(*header_size, allocator);
	if (*header == NULL)
		return LZMA_MEM_ERROR;

	// Magic
	memcpy(*header, lzma_header_magic, sizeof(lzma_header_magic));

	// Stream Flags
	stream_flags_encode(*header + sizeof(lzma_header_magic), options);

	// CRC32 of the Stream Header
	const uint32_t crc = lzma_crc32(
			*header + sizeof(lzma_header_magic), 1, 0);
	for (size_t i = 0; i < 4; ++i)
		(*header)[sizeof(lzma_header_magic) + 1 + i] = crc >> (i * 8);

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_stream_footer_encode(uint8_t **footer, size_t *footer_size,
		lzma_allocator *allocator, const lzma_stream_flags *options)
{
	// It's simpler to allocate a few bytes extra than too little.
	size_t alloc_size = sizeof(lzma_footer_magic) + 1 + LZMA_VLI_BYTES_MAX;
	if (!options->is_multi)
		*footer_size += LZMA_VLI_BYTES_MAX;

	*footer = lzma_alloc(alloc_size, allocator);
	if (*footer == NULL)
		return LZMA_MEM_ERROR;

	size_t pos = 0;

	// Uncompressed Size
	if (!options->is_multi && options->uncompressed_size != UINT64_MAX) {
		size_t dummy = 0;
		if (lzma_vli_encode(options->uncompressed_size,
				&dummy, 1, *footer, &pos, alloc_size)
				!= LZMA_STREAM_END)
			goto footer_error;
	}

	// Backward Size
	{
		// We must include the size of Uncompressed Size field.
		const uint64_t backward_size = options->backward_size + pos;

		size_t dummy = 0;
		if (lzma_vli_encode(backward_size, &dummy, 1,
				*footer, &pos, alloc_size) != LZMA_STREAM_END)
			goto footer_error;
	}

	// Stream Flags
	stream_flags_encode(*footer + pos++, options);

	// Magic
	memcpy(*footer + pos, lzma_footer_magic, sizeof(lzma_footer_magic));

	*footer_size = pos + sizeof(lzma_footer_magic);
	assert(*footer_size <= alloc_size);

	return LZMA_OK;

footer_error:
	free(*footer);
	return LZMA_HEADER_ERROR;
}
