///////////////////////////////////////////////////////////////////////////////
//
/// \file       subblock_encoder.c
/// \brief      Encoder of the Subblock filter
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "subblock_encoder.h"
#include "raw_encoder.h"


struct lzma_coder_s {
	lzma_next_coder next;
	bool end_was_reached;

	enum {
		MSEQ_RUN,
		MSEQ_FLUSH,
		MSEQ_SUBFILTER_INIT,
		MSEQ_SUBFILTER_UNCOMPRESSED_SIZE,
		MSEQ_SUBFILTER_FLAGS_INIT_1,
		MSEQ_SUBFILTER_FLAGS_INIT_2,
		MSEQ_SUBFILTER_FLAGS_COPY,
		MSEQ_SUBBLOCK_DATA_SIZE,

	} main_sequence;

	lzma_options_subblock *options;
	uint64_t uncompressed_size;

	size_t subblock_data_size;
	uint8_t *buffer;
	size_t buffer_pos;
	size_t buffer_size;
	enum {
		BSEQ_INIT,
		BSEQ_FILL,
		BSEQ_WRITE_SUBBLOCK_SIZE_0,
		BSEQ_WRITE_SUBBLOCK_SIZE_1,
		BSEQ_WRITE_SUBBLOCK_SIZE_2,
		BSEQ_WRITE_SUBBLOCK_SIZE_3,
		BSEQ_WRITE_DATA,
		BSEQ_WRITE_UNSET_MARKER,
	} buffer_sequence;

	lzma_next_coder subfilter;
	uint64_t subfilter_uncompressed_size;
	enum {
		SUB_NONE,
		SUB_UNCOMPRESSED_SIZE,
		SUB_END_OF_INPUT,
		SUB_END_OF_PAYLOAD_MARKER,
		SUB_UNSET,
	} subfilter_mode;

	size_t temp_pos;
	size_t temp_size;
	uint8_t temp[LZMA_BUFFER_SIZE];
};


static lzma_ret
subblock_buffer(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size)
	switch (coder->buffer_sequence) {
	case BSEQ_INIT: {
		if (*in_pos == in_size) {
			if (action == LZMA_FINISH
					&& coder->subfilter_mode == SUB_NONE)
				return LZMA_STREAM_END;

			return LZMA_OK;
		}

		coder->buffer_pos = 0;
		coder->buffer_size = 0;
		coder->buffer_sequence = BSEQ_FILL;
	}

	// Fall through

	case BSEQ_FILL:
		// Check that we have sane amount of input, so we can assume
		// it when updating uncompressed_size.
		if ((uint64_t)(in_size - *in_pos) > coder->uncompressed_size)
			return LZMA_DATA_ERROR;

		if (coder->subfilter_mode == SUB_NONE) {
			// No Subfilter is enabled, just copy
			// the data as is.
			const size_t in_used = bufcpy(
					in, in_pos, in_size, coder->buffer,
					&coder->buffer_size,
					coder->subblock_data_size);
			if (coder->uncompressed_size != UINT64_MAX)
				coder->uncompressed_size -= in_used;
		} else {
			// Store the current input position so that we can
			// update coder->uncompressed_size.
			const size_t in_start = *in_pos;

			// Call the Subfilter to fill the buffer.
			const lzma_ret ret = coder->subfilter.code(
					coder->subfilter.coder, allocator,
					in, in_pos, in_size,
					coder->buffer, &coder->buffer_size,
					coder->subblock_data_size, action);

			if (coder->uncompressed_size != UINT64_MAX)
				coder->uncompressed_size -= *in_pos - in_start;

			if (ret == LZMA_STREAM_END) {
				// FIXME?
				coder->options->subfilter.id = UINT64_MAX;

				if (coder->subfilter_mode == SUB_END_OF_INPUT)
					coder->subfilter_mode = SUB_UNSET;
				else
					coder->subfilter_mode = SUB_NONE;

			} else if (ret != LZMA_OK) {
				// NOTE: Keep in mind, that if
				// ret == LZMA_MEM_ERROR, it must
				// be possible to restart encoding.
				return ret;
			}
		}

		// If we ran out of input before the whole buffer
		// was filled, return to application.
		if (coder->buffer_size < coder->subblock_data_size
				&& action != LZMA_FINISH)
			return LZMA_OK;

		coder->buffer_sequence = BSEQ_WRITE_SUBBLOCK_SIZE_0;
		break;

	case BSEQ_WRITE_SUBBLOCK_SIZE_0:
		out[*out_pos] = coder->buffer_size & 0x0F;
		++*out_pos;
		coder->buffer_sequence = BSEQ_WRITE_SUBBLOCK_SIZE_1;
		break;

	case BSEQ_WRITE_SUBBLOCK_SIZE_1:
		out[*out_pos] = coder->buffer_size >> 4;
		++*out_pos;
		coder->buffer_sequence = BSEQ_WRITE_SUBBLOCK_SIZE_2;
		break;

	case BSEQ_WRITE_SUBBLOCK_SIZE_2:
		out[*out_pos] = coder->buffer_size >> 12;
		++*out_pos;
		coder->buffer_sequence = BSEQ_WRITE_SUBBLOCK_SIZE_3;
		break;

	case BSEQ_WRITE_SUBBLOCK_SIZE_3:
		out[*out_pos] = coder->buffer_size >> 20;
		++*out_pos;
		coder->buffer_sequence = BSEQ_WRITE_DATA;
		break;

	case BSEQ_WRITE_DATA:
		bufcpy(coder->buffer, &coder->buffer_pos, coder->buffer_size,
				out, out_pos, out_size);
		if (coder->buffer_pos < coder->buffer_size)
			return LZMA_OK;

		if (coder->subfilter_mode == SUB_UNSET)
			coder->buffer_sequence
					= BSEQ_WRITE_UNSET_MARKER;
		else
			coder->buffer_sequence = BSEQ_INIT;

		break;

	case BSEQ_WRITE_UNSET_MARKER:
		out[*out_pos] = 0x50;
		++*out_pos;
		coder->subfilter_mode = SUB_NONE;
		coder->buffer_sequence = BSEQ_INIT;
		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static lzma_ret
subblock_main(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size)
	switch (coder->main_sequence) {
	case MSEQ_RUN: {
		if (coder->options != NULL && coder->options->flush) {
			coder->main_sequence = MSEQ_FLUSH;
			break;
		}

		// When Subfilter filter uses known Uncompressed Size, we
		// need to limit the amount of data given to the Subfilter.
		// Once we have enough input data, we need also to pass
		// LZMA_FINISH as the action to make the Subfilter finish.
		size_t out_size_limit;
		lzma_action action_limit;
		if (coder->subfilter_mode == SUB_UNCOMPRESSED_SIZE
				&& (uint64_t)(out_size - *out_pos)
					> coder->subfilter_uncompressed_size) {
			out_size_limit = *out_pos + (size_t)(
					coder->subfilter_uncompressed_size);
			action_limit = LZMA_FINISH;
		} else {
			out_size_limit = out_size;
			action_limit = action;
		}

		const lzma_ret ret = subblock_buffer(coder, allocator,
					in, in_pos, in_size,
					out, out_pos, out_size, action);
		switch (ret) {
		case LZMA_OK:
			return LZMA_OK;

		case LZMA_STREAM_END:
			if (action == action_limit)
				return LZMA_STREAM_END;

			break;

		default:
			return ret;
		}

		break;
	}

	case MSEQ_FLUSH: {
		size_t dummy = 0;
		const lzma_ret ret = subblock_buffer(coder, allocator,
				NULL, &dummy, 0, out, out_pos, out_size,
				LZMA_FINISH);
		if (ret != LZMA_STREAM_END)
			return ret;

		coder->main_sequence = MSEQ_SUBFILTER_INIT;
		break;
	}

	case MSEQ_SUBFILTER_INIT:
		if (coder->options != NULL && coder->options->subfilter.id
				!= UINT64_MAX) {
			lzma_options_filter options[2];
			memcpy(options, &coder->options->subfilter,
					sizeof(coder->options->subfilter));
			options[1].id = UINT64_MAX;
			coder->subfilter_uncompressed_size = coder->options
					->subfilter_uncompressed_size;

			const lzma_ret ret = lzma_raw_encoder_init(
					&coder->subfilter, allocator,
					options,
					coder->subfilter_uncompressed_size,
					false);
			if (ret != LZMA_OK)
				return ret;

			if (coder->subfilter_uncompressed_size != UINT64_MAX) {
				out[*out_pos] = 0x30;
				coder->subfilter_mode = SUB_UNCOMPRESSED_SIZE;
				coder->main_sequence
					= MSEQ_SUBFILTER_UNCOMPRESSED_SIZE;
			} else {
				out[*out_pos] = 0x40;

				// Determine subfilter_mode i.e. if the
				// Subfilter needs unset marker or not.
				// Currently only LZMA uses EOPM.
				if (coder->options->subfilter.id
						== LZMA_FILTER_LZMA)
					coder->subfilter_mode
						= SUB_END_OF_PAYLOAD_MARKER;
				else
					coder->subfilter_mode
						= SUB_END_OF_INPUT;

				coder->main_sequence
					= MSEQ_SUBFILTER_FLAGS_INIT_1;
			}

			++*out_pos;
			break;
		}

		coder->main_sequence = MSEQ_SUBBLOCK_DATA_SIZE;
		break;

	case MSEQ_SUBFILTER_UNCOMPRESSED_SIZE: {
		const lzma_ret ret = lzma_vli_encode(
				coder->subfilter_uncompressed_size,
				&coder->buffer_pos, 1, out, out_pos, out_size);
				// FIXME ^ reset buffer_pos
		if (ret != LZMA_OK)
			return ret;

		coder->main_sequence = MSEQ_SUBFILTER_FLAGS_INIT_1;
		break;
	}

	case MSEQ_SUBFILTER_FLAGS_INIT_1: {
		coder->buffer_size = lzma_filter_flags_size(
				&coder->options->subfilter);

		// This shouldn't occur since we have already initialized
		// the Subfilter.
		if (coder->buffer_size == 0)
			return LZMA_STREAM_ERROR;

		lzma_free(coder->buffer, allocator);
		coder->buffer = lzma_alloc(coder->buffer_size, allocator);
		if (coder->buffer == NULL)
			return LZMA_MEM_ERROR;

		coder->main_sequence = MSEQ_SUBFILTER_FLAGS_INIT_2;
	}

	// Fall through

	case MSEQ_SUBFILTER_FLAGS_INIT_2: {
		size_t verify = 0;
		const lzma_ret ret = lzma_filter_flags_encode(coder->buffer,
				&verify, coder->buffer_size,
				&coder->options->subfilter);
		if (ret != LZMA_OK)
			return LZMA_HEADER_ERROR;

		coder->buffer_pos = 0;
		coder->main_sequence = MSEQ_SUBFILTER_FLAGS_COPY;
	}

	// Fall through

	case MSEQ_SUBFILTER_FLAGS_COPY: {
		bufcpy(coder->buffer, &coder->buffer_pos, coder->buffer_size,
				out, out_pos, out_size);
		if (coder->buffer_pos < coder->buffer_size)
			return LZMA_OK;

		lzma_free(coder->buffer, allocator);
		coder->buffer = NULL;
		coder->subblock_data_size = 0;

		coder->main_sequence = MSEQ_SUBBLOCK_DATA_SIZE;
		break;
	}

	case MSEQ_SUBBLOCK_DATA_SIZE: {
		size_t new_size;
		if (coder->options != NULL) {
			new_size = coder->options->subblock_data_size;
			if (new_size < LZMA_SUBBLOCK_DATA_SIZE_MIN
					|| new_size
						> LZMA_SUBBLOCK_DATA_SIZE_MAX)
				return LZMA_HEADER_ERROR;
		} else {
			new_size = LZMA_SUBBLOCK_DATA_SIZE_DEFAULT;
		}

		if (coder->subblock_data_size != new_size) {
			lzma_free(coder->buffer, allocator);
			coder->buffer = NULL;
			coder->subblock_data_size = 0;

			coder->buffer = lzma_alloc(new_size, allocator);
			if (coder->buffer == NULL)
				return LZMA_MEM_ERROR;

			coder->subblock_data_size = new_size;
		}

		if (coder->options != NULL)
			coder->options->flush = false;

		coder->main_sequence = MSEQ_RUN;
		break;
	}

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}



static lzma_ret
subblock_encode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	if (!coder->end_was_reached) {
		if (coder->next.code == NULL) {
			const lzma_ret ret = subblock_main(coder, allocator,
					in, in_pos, in_size,
					out, out_pos, out_size, action);
			if (ret != LZMA_STREAM_END)
				return ret;

			coder->end_was_reached = true;

		} else {
			do {
				if (coder->temp_pos == coder->temp_size) {
					coder->temp_pos = 0;
					coder->temp_size = 0;

					const lzma_ret ret = coder->next.code(
							coder->next.coder,
							allocator,
							in, in_pos, in_size,
							coder->temp,
							&coder->temp_size,
							LZMA_BUFFER_SIZE,
							action);
					if (ret == LZMA_STREAM_END) {
						assert(action == LZMA_FINISH);
						coder->end_was_reached = true;
					} else if (ret != LZMA_OK) {
						return ret;
					}
				}

				const lzma_ret ret = subblock_main(
						coder, allocator,
						coder->temp, &coder->temp_pos,
						coder->temp_size,
						out, out_pos, out_size,
						action);
				if (ret == LZMA_STREAM_END) {
					assert(action == LZMA_FINISH);
					assert(coder->end_was_reached);
					break;
				}

			} while (!coder->end_was_reached);
		}
	}

	// Write End of Payload Marker if uncompressed_size is unknown.
	if (coder->uncompressed_size == UINT64_MAX) {
		// If we have tough luck and there is not even one
		// byte of free space, return to application.
		if (*out_pos == out_size)
			return LZMA_OK;

		out[*out_pos] = 0x20;
		++*out_pos;

	} else if (coder->uncompressed_size != 0) {
		return LZMA_DATA_ERROR;
	}

	return LZMA_STREAM_END;
}


static void
subblock_encoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder->buffer, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_subblock_encoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_filter_info *filters)
{
	if (next->init != &lzma_subblock_encoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		next->coder->next = LZMA_NEXT_CODER_INIT;
		next->coder->buffer = NULL;
		next->coder->subfilter = LZMA_NEXT_CODER_INIT;
	}

	next->coder->end_was_reached = false;
	next->coder->main_sequence = MSEQ_SUBBLOCK_DATA_SIZE;
	next->coder->options = filters[0].options;
	next->coder->uncompressed_size = filters[0].uncompressed_size;
	next->coder->subblock_data_size = 0;
	next->coder->buffer_pos = 0;
	next->coder->buffer_size = 0;
	next->coder->buffer_sequence = BSEQ_INIT;
	next->coder->subfilter_mode = SUB_NONE;
	next->coder->temp_pos = 0;
	next->coder->temp_size = 0;

	{
		const lzma_ret ret = lzma_next_coder_init(&next->coder->next,
				allocator, filters + 1);
		if (ret != LZMA_OK) {
			subblock_encoder_end(next->coder, allocator);
			return ret;
		}
	}

	next->code = &subblock_encode;
	next->end = &subblock_encoder_end;

	return LZMA_OK;
}
